"""This module contains methods to featurize molecules.
Notes
-----
Input to all methods is a Molecule object.
Output of all methods is supplied with the method.

Supported output type
--------------------
'numpy': np.array
'rdkit': rdkit.DataStructs.cDataStructs.ExplicitBitVect

"""
import numpy as np
from rdkit.Chem import rdmolops
from rdkit import DataStructs
from rdkit.Chem import AllChem
from rdkit.Chem.Fingerprints import FingerprintMols
from rdkit.Chem import MACCSkeys
from rdkit.Chem.AtomPairs import Pairs, Torsions
from rdkit.DataStructs import cDataStructs
from mordred import Calculator, descriptors
from molSim.utils.ccbmlib_fingerprints import generate_fingerprints

from ..exceptions import (
    InvalidConfigurationError,
    NotInitializedError,
    MordredCalculatorError,
)


class Descriptor:
    """Class for descriptors.

    Attributes
    ----------
    label_: str
        Label used to denote the type of descriptor being used.
    numpy_: np.ndarray
        Value of the descriptor in the numpy format.
    rdkit_: rdkit.DataStructs.cDataStructs.UIntSparseIntVec
        Value of the descriptor in the rdkit format.

    """

    def __init__(self, value=None):
        if value is not None:
            self.set_manually(arbitrary_descriptor_val=value)

    def to_numpy(self):
        """Convert arbitrary fingerprints of type in_dtype to numpy arrays.

        Returns
        -------
        np.array

        """
        if self.check_init() is False:
            raise NotInitializedError(
                "Descriptor value not generated. Use "
                "make_fingerprint() to initialize it."
            )
        if not hasattr(self, "numpy_"):
            self.numpy_ = np.zeros((0,), dtype=np.int8)
            DataStructs.ConvertToNumpyArray(self.rdkit_, self.numpy_)
        self.numpy_ = self.numpy_.flatten()
        return self.numpy_

    def to_rdkit(self):
        if self.check_init() is False:
            raise NotInitializedError(
                "Descriptor value not generated. Use "
                "make_fingerprint() to initialize it."
            )
        if not hasattr(self, "rdkit_"):
            raise ValueError(
                "Attempting to convert arbitrary numpy array "
                "to rdkit bit vector is not supported"
            )
        return self.rdkit_

    def check_init(self):
        if hasattr(self, "numpy_") or hasattr(self, "rdkit_"):
            return True
        return False

    def _set_morgan_fingerprint(self, molecule_graph, radius, n_bits, **kwargs):
        """Set the descriptor to a morgan fingerprint.

        Parameters
        ----------
        molecule_graph: RDKIT object
            Graph of molecule to be fingerprinted.
        radius: int
            Radius of fingerprint, 3 corresponds to diameter (ECFP)6.
        n_bits: int
            Number of bits to use if Morgan Fingerprint wanted as
            a bit vector. If set to None, Morgan fingerprint returned
            as count.

        """
        self.rdkit_ = AllChem.GetMorganFingerprintAsBitVect(
            molecule_graph, radius, nBits=n_bits
        )
        self.label_ = "morgan_fingerprint"
        self.params_ = {"radius": radius, "n_bits": n_bits}

    def _set_rdkit_topological_fingerprint(
        self, molecule_graph, min_path, max_path, **kwargs
    ):
        """Set the descriptor to a topological fingerprint.

        Parameters
        ----------
        molecule_graph: RDKIT object
            Graph of molecule to be fingerprinted.
        min_path: int
            Minimum path used to generate the topological fingerprint.
        max_path: int
            Maximum path used to generate the topological fingerprint.

        """
        if molecule_graph.GetNumAtoms() <= min_path:
            raise InvalidConfigurationError(
                f"# atoms in molecule: "
                f"{molecule_graph.GetNumAtoms()}, "
                f"min_path: {min_path}. "
                f"For topological fingerprint, "
                f"the number of atoms has to be "
                f"greater than the minimum path "
                f"used for fingerprint."
            )
        self.rdkit_ = rdmolops.RDKFingerprint(
            molecule_graph, minPath=min_path, maxPath=max_path
        )
        self.label_ = "topological_fingerprint"
        self.params_ = {"min_path": min_path, "max_path": max_path}

    def _set_mordred_descriptor(self, molecule_graph, descriptor, **kwargs):
        """Set the value of numpy_ to the descriptor as indicated by descriptor.

        Args:
            molecule_graph (RDKit object): Graph of the molecule of interest.
            descriptor (string): Name of descriptor, as implemented in Mordred.

        Raises:
            MordredCalculatorError: If Morded is unable to calculate a value
                for the indicated descriptor, this exception will be raised.
        """
        try:
            calc = Calculator(descriptors, ignore_3D=False)
            res = calc(molecule_graph)
            res.drop_missing()
            self.numpy_ = np.array(res[descriptor])
            self.label_ = descriptor
        except KeyError:
            raise MordredCalculatorError(
                """Mordred descriptor calculator unable to calculate descriptor \"{}\",
                ensure correct name is used (https://mordred-descriptor.github.io/documentation/master/descriptors.html).""".format(
                    descriptor
                )
            )

    def _set_daylight_fingerprint(self, molecule_graph, **kwargs):
        """Set the descriptor to a daylight fingerprint.

        Parameters
        ----------
        molecule_graph: RDKIT object
            Graph of molecule to be fingerprinted.

        """
        # returns an ExplicitBitVect
        self.rdkit_ = FingerprintMols.FingerprintMol(molecule_graph)
        self.label_ = "daylight_fingerprint"
        self.params_ = {}

    def _set_maccs_keys(self, molecule_graph, **kwargs):
        """Set the descriptor to MACCS keys.

        Parameters
        ----------
        molecule_graph: RDKIT object
            Graph of molecule to be fingerprinted.

        """
        # returns an ExplicitBitVect
        self.rdkit_ = MACCSkeys.GenMACCSKeys(molecule_graph)
        self.label_ = "maccs_keys"
        self.params_ = {}

    def _set_atom_pair_fingerprint(self, molecule_graph, **kwargs):
        """Set the descriptor to an atom-pair fingerprint.

        Parameters
        ----------
        molecule_graph: RDKIT object
            Graph of molecule to be fingerprinted.

        """
        # returns a SparseBitVect
        fp = Pairs.GetAtomPairFingerprintAsBitVect(molecule_graph)
        fp2 = cDataStructs.ConvertToExplicit(fp)
        self.rdkit_ = fp2
        self.label_ = "atom-pair_fingerprint"
        self.params_ = {}

    def _set_torsion_fingerprint(self, molecule_graph, **kwargs):
        """Set the descriptor to a torsion fingerprint.

        Parameters
        ----------
        molecule_graph: RDKIT object
            Graph of molecule to be fingerprinted.

        """
        # returns a long sparse int vector
        fp = Torsions.GetTopologicalTorsionFingerprintAsIntVect(molecule_graph)
        self.rdkit_ = fp
        self.label_ = "torsion_fingerprint"
        self.params_ = {}

    def _set_ccbmlib_fingerprint(self, molecule_graph, descriptor, **kwargs):
        """Set the descriptor to fingerprint from ccbmlib.

        Parameters
        ----------
        molecule_graph: RDKIT object
            Graph of molecule to be fingerprinted.

        """
        # returns a list of ints which represent the on bits (features)
        fp = generate_fingerprints(molecule_graph, descriptor, **kwargs)
        self.numpy_ = fp
        self.label_ = descriptor
        self.params_ = {}

    def make_fingerprint(
        self, molecule_graph, fingerprint_type, fingerprint_params=None
    ):
        """Make fingerprint of a molecule based on a graph representation.
        Set the state of the descriptor to this fingerprint.

        Args:
            molecule_graph (RDKIT object): The graph object used to make a
                fingerprint.
            fingerprint_type (str): label for the type of fingerprint.
                Invokes get_supported_descriptors()['fingerprints']
                for list of supported fingerprints.
            fingerprint_params (dict): Keyword arguments used to modify
                parameters of fingerprint. Default is None.
        """
        if fingerprint_params is None:
            fingerprint_params = {}
        if fingerprint_type == "morgan_fingerprint":
            morgan_params = {"radius": 3, "n_bits": 1024}
            morgan_params.update(fingerprint_params)
            self._set_morgan_fingerprint(molecule_graph=molecule_graph, **morgan_params)
        elif fingerprint_type == "topological_fingerprint":
            topological_params = {"min_path": 1, "max_path": 7}
            topological_params.update(fingerprint_params)
            self._set_rdkit_topological_fingerprint(
                molecule_graph=molecule_graph, **topological_params
            )
        elif fingerprint_type == "daylight_fingerprint":
            daylight_params = {}
            daylight_params.update(fingerprint_params)
            self._set_daylight_fingerprint(
                molecule_graph=molecule_graph, **daylight_params
            )
        elif fingerprint_type == "maccs_keys":
            maccs_params = {}
            maccs_params.update(fingerprint_params)
            self._set_maccs_keys(molecule_graph=molecule_graph, **maccs_params)
        elif fingerprint_type == "atom-pair_fingerprint":
            atom_pair_fp_params = {}
            atom_pair_fp_params.update(fingerprint_params)
            self._set_atom_pair_fingerprint(
                molecule_graph=molecule_graph, **atom_pair_fp_params
            )
        elif fingerprint_type == "torsion_fingerprint":
            torsion_params = {}
            torsion_params.update(fingerprint_params)
            self._set_torsion_fingerprint(
                molecule_graph=molecule_graph, **torsion_params
            )
        elif fingerprint_type.split(":")[0] == "mordred":
            mordred_params = {}
            self._set_mordred_descriptor(
                molecule_graph=molecule_graph,
                descriptor=fingerprint_type.split(":")[1],
                **mordred_params,
            )
        elif fingerprint_type.split(":")[0] == "ccbmlib":
            ccbmlib_params = {}
            self._set_ccbmlib_fingerprint(
                molecule_graph=molecule_graph,
                descriptor=fingerprint_type.split(":")[1],
                **ccbmlib_params,
            )
        else:
            raise ValueError(f"{fingerprint_type} not supported")

    def set_manually(self, arbitrary_descriptor_val):
        """Set the descriptor value manually based on user specified value.

        Args:
            arbitrary_descriptor_val (np.ndarray or list): Vectorized
                representation of descriptor values.
        """
        self.label_ = "arbitrary"
        self.numpy_ = np.array(arbitrary_descriptor_val)

    def get_label(self):
        if not self.check_init():
            raise NotInitializedError
        else:
            return self.label_

    def get_params(self):
        if not self.check_init():
            raise NotInitializedError
        else:
            return self.params_

    def is_fingerprint(self):
        return "fingerprint" in self.get_label()

    def get_folded_fprint(self, fold_to_length):
        """
        Get the folded value of a fingerprint to a specified length.
        Args:
            fold_to_length (int): Number of bits to fold to.

        Returns:
            (np.ndarray): Folded fingerprint.
        """
        if not self.is_fingerprint():
            raise ValueError("Can only fold fingerprints")
        fingerprint = self.to_numpy()
        if len(fingerprint) < fold_to_length:
            raise InvalidConfigurationError(
                f"Cannot fold fingerprint of "
                f"length {len(fingerprint)}to a "
                f"higher length {fold_to_length}"
            )
        n_folds = np.log2(len(fingerprint) / fold_to_length)
        if n_folds - int(n_folds) > 0.0:
            raise InvalidConfigurationError(
                f"Fingerprint length "
                f"{len(fingerprint)} not "
                f"a 2-multiple of required "
                f"folded length {fold_to_length}"
            )
        for _ in range(int(n_folds)):
            mid_point = int(len(fingerprint) / 2)
            assert mid_point - (len(fingerprint) / 2) == 0.0
            fingerprint = fingerprint[:mid_point] | fingerprint[mid_point:]
        assert len(fingerprint) == fold_to_length
        return fingerprint

    @staticmethod
    def shorten_label(label):
        """Shorten the label of a fingerprint. Useful for plotting purposes.

        Args:
            label (str): Label of fingerprint to shorten.

        Returns:
            (str): Shortened label.

        Raises:
            InvalidConfigurationError: if label not in
                get_supported_descriptors()

        Currently implemented shortening strategies:
            1. Fingerprints: remove '_fingerprint' from the label

        """
        if label not in Descriptor.get_all_supported_descriptors():
            raise InvalidConfigurationError(f"{label} not a " f"supported descriptor")
        if label in Descriptor.get_supported_fprints():
            return label.replace("_fingerprint", "")
        return label

    @staticmethod
    def fold_to_equal_length(fingerprint1, fingerprint2):
        """
        Get back two fingerprint arrays of equal length. The longer fingerprint
        is folded to the size of the smaller one.

        Args:
            fingerprint1 (Descriptor): Fingerprint one.
            fingerprint2 (Descriptor): Fingerprint two.

        Returns:
            (np.ndarray, np.ndarray)

        """
        fprint1_arr = fingerprint1.to_numpy()
        fprint2_arr = fingerprint2.to_numpy()
        if len(fprint1_arr) > len(fprint2_arr):
            return (
                fingerprint1.get_folded_fprint(fold_to_length=len(fprint2_arr)),
                fprint2_arr,
            )
        else:
            return fprint1_arr, fingerprint2.get_folded_fprint(
                fold_to_length=len(fprint1_arr)
            )

    @staticmethod
    def get_supported_fprints():
        """Return a list of strings for the currently implemented
        molecular fingerprints.
        Returns:
            List: List of strings.
        """
        return [
            "morgan_fingerprint",
            "topological_fingerprint",
            "daylight_fingerprint",
        ]

    @staticmethod
    def get_all_supported_descriptors():
        """Returns a list of descriptors which _can_ be used with molSim
        but are considered experimental or are complex/rarely used and
        are therefore excluded from the UI.

        Returns:
            List: List of strings of all supported descriptors.
        """
        return [
            "morgan_fingerprint",
            "topological_fingerprint",
            "daylight_fingerprint",
            "maccs_keys",
            "atom-pair_fingerprint",
            "torsion_fingerprint",
            "ccbmlib:atom_pairs",
            "ccbmlib:hashed_atom_pairs",
            "ccbmlib:avalon",
            "ccbmlib:maccs_keys",
            "ccbmlib:morgan",
            "ccbmlib:hashed_morgan",
            "ccbmlib:rdkit_fingerprint",
            "ccbmlib:torsions",
            "ccbmlib:hashed_torsions",
            "mordred:ABC",
            "mordred:ABCGG",
            "mordred:nAcid",
            "mordred:nBase",
            "mordred:SpAbs_A",
            "mordred:SpMax_A",
            "mordred:SpDiam_A",
            "mordred:SpAD_A",
            "mordred:SpMAD_A",
            "mordred:LogEE_A",
            "mordred:VE1_A",
            "mordred:VE2_A",
            "mordred:VE3_A",
            "mordred:VR1_A",
            "mordred:VR2_A",
            "mordred:VR3_A",
            "mordred:nAromAtom",
            "mordred:nAromBond",
            "mordred:nAtom",
            "mordred:nHeavyAtom",
            "mordred:nSpiro",
            "mordred:nBridgehead",
            "mordred:nHetero",
            "mordred:nH",
            "mordred:nB",
            "mordred:nC",
            "mordred:nN",
            "mordred:nO",
            "mordred:nS",
            "mordred:nP",
            "mordred:nF",
            "mordred:nCl",
            "mordred:nBr",
            "mordred:nI",
            "mordred:nX",
            "mordred:ATS0dv",
            "mordred:ATS1dv",
            "mordred:ATS2dv",
            "mordred:ATS3dv",
            "mordred:ATS4dv",
            "mordred:ATS5dv",
            "mordred:ATS6dv",
            "mordred:ATS7dv",
            "mordred:ATS8dv",
            "mordred:ATS0d",
            "mordred:ATS1d",
            "mordred:ATS2d",
            "mordred:ATS3d",
            "mordred:ATS4d",
            "mordred:ATS5d",
            "mordred:ATS6d",
            "mordred:ATS7d",
            "mordred:ATS8d",
            "mordred:ATS0s",
            "mordred:ATS1s",
            "mordred:ATS2s",
            "mordred:ATS3s",
            "mordred:ATS4s",
            "mordred:ATS5s",
            "mordred:ATS6s",
            "mordred:ATS7s",
            "mordred:ATS8s",
            "mordred:ATS0Z",
            "mordred:ATS1Z",
            "mordred:ATS2Z",
            "mordred:ATS3Z",
            "mordred:ATS4Z",
            "mordred:ATS5Z",
            "mordred:ATS6Z",
            "mordred:ATS7Z",
            "mordred:ATS8Z",
            "mordred:ATS0m",
            "mordred:ATS1m",
            "mordred:ATS2m",
            "mordred:ATS3m",
            "mordred:ATS4m",
            "mordred:ATS5m",
            "mordred:ATS6m",
            "mordred:ATS7m",
            "mordred:ATS8m",
            "mordred:ATS0v",
            "mordred:ATS1v",
            "mordred:ATS2v",
            "mordred:ATS3v",
            "mordred:ATS4v",
            "mordred:ATS5v",
            "mordred:ATS6v",
            "mordred:ATS7v",
            "mordred:ATS8v",
            "mordred:ATS0se",
            "mordred:ATS1se",
            "mordred:ATS2se",
            "mordred:ATS3se",
            "mordred:ATS4se",
            "mordred:ATS5se",
            "mordred:ATS6se",
            "mordred:ATS7se",
            "mordred:ATS8se",
            "mordred:ATS0pe",
            "mordred:ATS1pe",
            "mordred:ATS2pe",
            "mordred:ATS3pe",
            "mordred:ATS4pe",
            "mordred:ATS5pe",
            "mordred:ATS6pe",
            "mordred:ATS7pe",
            "mordred:ATS8pe",
            "mordred:ATS0are",
            "mordred:ATS1are",
            "mordred:ATS2are",
            "mordred:ATS3are",
            "mordred:ATS4are",
            "mordred:ATS5are",
            "mordred:ATS6are",
            "mordred:ATS7are",
            "mordred:ATS8are",
            "mordred:ATS0p",
            "mordred:ATS1p",
            "mordred:ATS2p",
            "mordred:ATS3p",
            "mordred:ATS4p",
            "mordred:ATS5p",
            "mordred:ATS6p",
            "mordred:ATS7p",
            "mordred:ATS8p",
            "mordred:ATS0i",
            "mordred:ATS1i",
            "mordred:ATS2i",
            "mordred:ATS3i",
            "mordred:ATS4i",
            "mordred:ATS5i",
            "mordred:ATS6i",
            "mordred:ATS7i",
            "mordred:ATS8i",
            "mordred:AATS0dv",
            "mordred:AATS1dv",
            "mordred:AATS2dv",
            "mordred:AATS3dv",
            "mordred:AATS4dv",
            "mordred:AATS5dv",
            "mordred:AATS6dv",
            "mordred:AATS7dv",
            "mordred:AATS8dv",
            "mordred:AATS0d",
            "mordred:AATS1d",
            "mordred:AATS2d",
            "mordred:AATS3d",
            "mordred:AATS4d",
            "mordred:AATS5d",
            "mordred:AATS6d",
            "mordred:AATS7d",
            "mordred:AATS8d",
            "mordred:AATS0s",
            "mordred:AATS1s",
            "mordred:AATS2s",
            "mordred:AATS3s",
            "mordred:AATS4s",
            "mordred:AATS5s",
            "mordred:AATS6s",
            "mordred:AATS7s",
            "mordred:AATS8s",
            "mordred:AATS0Z",
            "mordred:AATS1Z",
            "mordred:AATS2Z",
            "mordred:AATS3Z",
            "mordred:AATS4Z",
            "mordred:AATS5Z",
            "mordred:AATS6Z",
            "mordred:AATS7Z",
            "mordred:AATS8Z",
            "mordred:AATS0m",
            "mordred:AATS1m",
            "mordred:AATS2m",
            "mordred:AATS3m",
            "mordred:AATS4m",
            "mordred:AATS5m",
            "mordred:AATS6m",
            "mordred:AATS7m",
            "mordred:AATS8m",
            "mordred:AATS0v",
            "mordred:AATS1v",
            "mordred:AATS2v",
            "mordred:AATS3v",
            "mordred:AATS4v",
            "mordred:AATS5v",
            "mordred:AATS6v",
            "mordred:AATS7v",
            "mordred:AATS8v",
            "mordred:AATS0se",
            "mordred:AATS1se",
            "mordred:AATS2se",
            "mordred:AATS3se",
            "mordred:AATS4se",
            "mordred:AATS5se",
            "mordred:AATS6se",
            "mordred:AATS7se",
            "mordred:AATS8se",
            "mordred:AATS0pe",
            "mordred:AATS1pe",
            "mordred:AATS2pe",
            "mordred:AATS3pe",
            "mordred:AATS4pe",
            "mordred:AATS5pe",
            "mordred:AATS6pe",
            "mordred:AATS7pe",
            "mordred:AATS8pe",
            "mordred:AATS0are",
            "mordred:AATS1are",
            "mordred:AATS2are",
            "mordred:AATS3are",
            "mordred:AATS4are",
            "mordred:AATS5are",
            "mordred:AATS6are",
            "mordred:AATS7are",
            "mordred:AATS8are",
            "mordred:AATS0p",
            "mordred:AATS1p",
            "mordred:AATS2p",
            "mordred:AATS3p",
            "mordred:AATS4p",
            "mordred:AATS5p",
            "mordred:AATS6p",
            "mordred:AATS7p",
            "mordred:AATS8p",
            "mordred:AATS0i",
            "mordred:AATS1i",
            "mordred:AATS2i",
            "mordred:AATS3i",
            "mordred:AATS4i",
            "mordred:AATS5i",
            "mordred:AATS6i",
            "mordred:AATS7i",
            "mordred:AATS8i",
            "mordred:ATSC0c",
            "mordred:ATSC1c",
            "mordred:ATSC2c",
            "mordred:ATSC3c",
            "mordred:ATSC4c",
            "mordred:ATSC5c",
            "mordred:ATSC6c",
            "mordred:ATSC7c",
            "mordred:ATSC8c",
            "mordred:ATSC0dv",
            "mordred:ATSC1dv",
            "mordred:ATSC2dv",
            "mordred:ATSC3dv",
            "mordred:ATSC4dv",
            "mordred:ATSC5dv",
            "mordred:ATSC6dv",
            "mordred:ATSC7dv",
            "mordred:ATSC8dv",
            "mordred:ATSC0d",
            "mordred:ATSC1d",
            "mordred:ATSC2d",
            "mordred:ATSC3d",
            "mordred:ATSC4d",
            "mordred:ATSC5d",
            "mordred:ATSC6d",
            "mordred:ATSC7d",
            "mordred:ATSC8d",
            "mordred:ATSC0s",
            "mordred:ATSC1s",
            "mordred:ATSC2s",
            "mordred:ATSC3s",
            "mordred:ATSC4s",
            "mordred:ATSC5s",
            "mordred:ATSC6s",
            "mordred:ATSC7s",
            "mordred:ATSC8s",
            "mordred:ATSC0Z",
            "mordred:ATSC1Z",
            "mordred:ATSC2Z",
            "mordred:ATSC3Z",
            "mordred:ATSC4Z",
            "mordred:ATSC5Z",
            "mordred:ATSC6Z",
            "mordred:ATSC7Z",
            "mordred:ATSC8Z",
            "mordred:ATSC0m",
            "mordred:ATSC1m",
            "mordred:ATSC2m",
            "mordred:ATSC3m",
            "mordred:ATSC4m",
            "mordred:ATSC5m",
            "mordred:ATSC6m",
            "mordred:ATSC7m",
            "mordred:ATSC8m",
            "mordred:ATSC0v",
            "mordred:ATSC1v",
            "mordred:ATSC2v",
            "mordred:ATSC3v",
            "mordred:ATSC4v",
            "mordred:ATSC5v",
            "mordred:ATSC6v",
            "mordred:ATSC7v",
            "mordred:ATSC8v",
            "mordred:ATSC0se",
            "mordred:ATSC1se",
            "mordred:ATSC2se",
            "mordred:ATSC3se",
            "mordred:ATSC4se",
            "mordred:ATSC5se",
            "mordred:ATSC6se",
            "mordred:ATSC7se",
            "mordred:ATSC8se",
            "mordred:ATSC0pe",
            "mordred:ATSC1pe",
            "mordred:ATSC2pe",
            "mordred:ATSC3pe",
            "mordred:ATSC4pe",
            "mordred:ATSC5pe",
            "mordred:ATSC6pe",
            "mordred:ATSC7pe",
            "mordred:ATSC8pe",
            "mordred:ATSC0are",
            "mordred:ATSC1are",
            "mordred:ATSC2are",
            "mordred:ATSC3are",
            "mordred:ATSC4are",
            "mordred:ATSC5are",
            "mordred:ATSC6are",
            "mordred:ATSC7are",
            "mordred:ATSC8are",
            "mordred:ATSC0p",
            "mordred:ATSC1p",
            "mordred:ATSC2p",
            "mordred:ATSC3p",
            "mordred:ATSC4p",
            "mordred:ATSC5p",
            "mordred:ATSC6p",
            "mordred:ATSC7p",
            "mordred:ATSC8p",
            "mordred:ATSC0i",
            "mordred:ATSC1i",
            "mordred:ATSC2i",
            "mordred:ATSC3i",
            "mordred:ATSC4i",
            "mordred:ATSC5i",
            "mordred:ATSC6i",
            "mordred:ATSC7i",
            "mordred:ATSC8i",
            "mordred:AATSC0c",
            "mordred:AATSC1c",
            "mordred:AATSC2c",
            "mordred:AATSC3c",
            "mordred:AATSC4c",
            "mordred:AATSC5c",
            "mordred:AATSC6c",
            "mordred:AATSC7c",
            "mordred:AATSC8c",
            "mordred:AATSC0dv",
            "mordred:AATSC1dv",
            "mordred:AATSC2dv",
            "mordred:AATSC3dv",
            "mordred:AATSC4dv",
            "mordred:AATSC5dv",
            "mordred:AATSC6dv",
            "mordred:AATSC7dv",
            "mordred:AATSC8dv",
            "mordred:AATSC0d",
            "mordred:AATSC1d",
            "mordred:AATSC2d",
            "mordred:AATSC3d",
            "mordred:AATSC4d",
            "mordred:AATSC5d",
            "mordred:AATSC6d",
            "mordred:AATSC7d",
            "mordred:AATSC8d",
            "mordred:AATSC0s",
            "mordred:AATSC1s",
            "mordred:AATSC2s",
            "mordred:AATSC3s",
            "mordred:AATSC4s",
            "mordred:AATSC5s",
            "mordred:AATSC6s",
            "mordred:AATSC7s",
            "mordred:AATSC8s",
            "mordred:AATSC0Z",
            "mordred:AATSC1Z",
            "mordred:AATSC2Z",
            "mordred:AATSC3Z",
            "mordred:AATSC4Z",
            "mordred:AATSC5Z",
            "mordred:AATSC6Z",
            "mordred:AATSC7Z",
            "mordred:AATSC8Z",
            "mordred:AATSC0m",
            "mordred:AATSC1m",
            "mordred:AATSC2m",
            "mordred:AATSC3m",
            "mordred:AATSC4m",
            "mordred:AATSC5m",
            "mordred:AATSC6m",
            "mordred:AATSC7m",
            "mordred:AATSC8m",
            "mordred:AATSC0v",
            "mordred:AATSC1v",
            "mordred:AATSC2v",
            "mordred:AATSC3v",
            "mordred:AATSC4v",
            "mordred:AATSC5v",
            "mordred:AATSC6v",
            "mordred:AATSC7v",
            "mordred:AATSC8v",
            "mordred:AATSC0se",
            "mordred:AATSC1se",
            "mordred:AATSC2se",
            "mordred:AATSC3se",
            "mordred:AATSC4se",
            "mordred:AATSC5se",
            "mordred:AATSC6se",
            "mordred:AATSC7se",
            "mordred:AATSC8se",
            "mordred:AATSC0pe",
            "mordred:AATSC1pe",
            "mordred:AATSC2pe",
            "mordred:AATSC3pe",
            "mordred:AATSC4pe",
            "mordred:AATSC5pe",
            "mordred:AATSC6pe",
            "mordred:AATSC7pe",
            "mordred:AATSC8pe",
            "mordred:AATSC0are",
            "mordred:AATSC1are",
            "mordred:AATSC2are",
            "mordred:AATSC3are",
            "mordred:AATSC4are",
            "mordred:AATSC5are",
            "mordred:AATSC6are",
            "mordred:AATSC7are",
            "mordred:AATSC8are",
            "mordred:AATSC0p",
            "mordred:AATSC1p",
            "mordred:AATSC2p",
            "mordred:AATSC3p",
            "mordred:AATSC4p",
            "mordred:AATSC5p",
            "mordred:AATSC6p",
            "mordred:AATSC7p",
            "mordred:AATSC8p",
            "mordred:AATSC0i",
            "mordred:AATSC1i",
            "mordred:AATSC2i",
            "mordred:AATSC3i",
            "mordred:AATSC4i",
            "mordred:AATSC5i",
            "mordred:AATSC6i",
            "mordred:AATSC7i",
            "mordred:AATSC8i",
            "mordred:MATS1c",
            "mordred:MATS2c",
            "mordred:MATS3c",
            "mordred:MATS4c",
            "mordred:MATS5c",
            "mordred:MATS6c",
            "mordred:MATS7c",
            "mordred:MATS8c",
            "mordred:MATS1dv",
            "mordred:MATS2dv",
            "mordred:MATS3dv",
            "mordred:MATS4dv",
            "mordred:MATS5dv",
            "mordred:MATS6dv",
            "mordred:MATS7dv",
            "mordred:MATS8dv",
            "mordred:MATS1d",
            "mordred:MATS2d",
            "mordred:MATS3d",
            "mordred:MATS4d",
            "mordred:MATS5d",
            "mordred:MATS6d",
            "mordred:MATS7d",
            "mordred:MATS8d",
            "mordred:MATS1s",
            "mordred:MATS2s",
            "mordred:MATS3s",
            "mordred:MATS4s",
            "mordred:MATS5s",
            "mordred:MATS6s",
            "mordred:MATS7s",
            "mordred:MATS8s",
            "mordred:MATS1Z",
            "mordred:MATS2Z",
            "mordred:MATS3Z",
            "mordred:MATS4Z",
            "mordred:MATS5Z",
            "mordred:MATS6Z",
            "mordred:MATS7Z",
            "mordred:MATS8Z",
            "mordred:MATS1m",
            "mordred:MATS2m",
            "mordred:MATS3m",
            "mordred:MATS4m",
            "mordred:MATS5m",
            "mordred:MATS6m",
            "mordred:MATS7m",
            "mordred:MATS8m",
            "mordred:MATS1v",
            "mordred:MATS2v",
            "mordred:MATS3v",
            "mordred:MATS4v",
            "mordred:MATS5v",
            "mordred:MATS6v",
            "mordred:MATS7v",
            "mordred:MATS8v",
            "mordred:MATS1se",
            "mordred:MATS2se",
            "mordred:MATS3se",
            "mordred:MATS4se",
            "mordred:MATS5se",
            "mordred:MATS6se",
            "mordred:MATS7se",
            "mordred:MATS8se",
            "mordred:MATS1pe",
            "mordred:MATS2pe",
            "mordred:MATS3pe",
            "mordred:MATS4pe",
            "mordred:MATS5pe",
            "mordred:MATS6pe",
            "mordred:MATS7pe",
            "mordred:MATS8pe",
            "mordred:MATS1are",
            "mordred:MATS2are",
            "mordred:MATS3are",
            "mordred:MATS4are",
            "mordred:MATS5are",
            "mordred:MATS6are",
            "mordred:MATS7are",
            "mordred:MATS8are",
            "mordred:MATS1p",
            "mordred:MATS2p",
            "mordred:MATS3p",
            "mordred:MATS4p",
            "mordred:MATS5p",
            "mordred:MATS6p",
            "mordred:MATS7p",
            "mordred:MATS8p",
            "mordred:MATS1i",
            "mordred:MATS2i",
            "mordred:MATS3i",
            "mordred:MATS4i",
            "mordred:MATS5i",
            "mordred:MATS6i",
            "mordred:MATS7i",
            "mordred:MATS8i",
            "mordred:GATS1c",
            "mordred:GATS2c",
            "mordred:GATS3c",
            "mordred:GATS4c",
            "mordred:GATS5c",
            "mordred:GATS6c",
            "mordred:GATS7c",
            "mordred:GATS8c",
            "mordred:GATS1dv",
            "mordred:GATS2dv",
            "mordred:GATS3dv",
            "mordred:GATS4dv",
            "mordred:GATS5dv",
            "mordred:GATS6dv",
            "mordred:GATS7dv",
            "mordred:GATS8dv",
            "mordred:GATS1d",
            "mordred:GATS2d",
            "mordred:GATS3d",
            "mordred:GATS4d",
            "mordred:GATS5d",
            "mordred:GATS6d",
            "mordred:GATS7d",
            "mordred:GATS8d",
            "mordred:GATS1s",
            "mordred:GATS2s",
            "mordred:GATS3s",
            "mordred:GATS4s",
            "mordred:GATS5s",
            "mordred:GATS6s",
            "mordred:GATS7s",
            "mordred:GATS8s",
            "mordred:GATS1Z",
            "mordred:GATS2Z",
            "mordred:GATS3Z",
            "mordred:GATS4Z",
            "mordred:GATS5Z",
            "mordred:GATS6Z",
            "mordred:GATS7Z",
            "mordred:GATS8Z",
            "mordred:GATS1m",
            "mordred:GATS2m",
            "mordred:GATS3m",
            "mordred:GATS4m",
            "mordred:GATS5m",
            "mordred:GATS6m",
            "mordred:GATS7m",
            "mordred:GATS8m",
            "mordred:GATS1v",
            "mordred:GATS2v",
            "mordred:GATS3v",
            "mordred:GATS4v",
            "mordred:GATS5v",
            "mordred:GATS6v",
            "mordred:GATS7v",
            "mordred:GATS8v",
            "mordred:GATS1se",
            "mordred:GATS2se",
            "mordred:GATS3se",
            "mordred:GATS4se",
            "mordred:GATS5se",
            "mordred:GATS6se",
            "mordred:GATS7se",
            "mordred:GATS8se",
            "mordred:GATS1pe",
            "mordred:GATS2pe",
            "mordred:GATS3pe",
            "mordred:GATS4pe",
            "mordred:GATS5pe",
            "mordred:GATS6pe",
            "mordred:GATS7pe",
            "mordred:GATS8pe",
            "mordred:GATS1are",
            "mordred:GATS2are",
            "mordred:GATS3are",
            "mordred:GATS4are",
            "mordred:GATS5are",
            "mordred:GATS6are",
            "mordred:GATS7are",
            "mordred:GATS8are",
            "mordred:GATS1p",
            "mordred:GATS2p",
            "mordred:GATS3p",
            "mordred:GATS4p",
            "mordred:GATS5p",
            "mordred:GATS6p",
            "mordred:GATS7p",
            "mordred:GATS8p",
            "mordred:GATS1i",
            "mordred:GATS2i",
            "mordred:GATS3i",
            "mordred:GATS4i",
            "mordred:GATS5i",
            "mordred:GATS6i",
            "mordred:GATS7i",
            "mordred:GATS8i",
            "mordred:BCUTc-1h",
            "mordred:BCUTc-1l",
            "mordred:BCUTdv-1h",
            "mordred:BCUTdv-1l",
            "mordred:BCUTd-1h",
            "mordred:BCUTd-1l",
            "mordred:BCUTs-1h",
            "mordred:BCUTs-1l",
            "mordred:BCUTZ-1h",
            "mordred:BCUTZ-1l",
            "mordred:BCUTm-1h",
            "mordred:BCUTm-1l",
            "mordred:BCUTv-1h",
            "mordred:BCUTv-1l",
            "mordred:BCUTse-1h",
            "mordred:BCUTse-1l",
            "mordred:BCUTpe-1h",
            "mordred:BCUTpe-1l",
            "mordred:BCUTare-1h",
            "mordred:BCUTare-1l",
            "mordred:BCUTp-1h",
            "mordred:BCUTp-1l",
            "mordred:BCUTi-1h",
            "mordred:BCUTi-1l",
            "mordred:BalabanJ",
            "mordred:SpAbs_DzZ",
            "mordred:SpMax_DzZ",
            "mordred:SpDiam_DzZ",
            "mordred:SpAD_DzZ",
            "mordred:SpMAD_DzZ",
            "mordred:LogEE_DzZ",
            "mordred:SM1_DzZ",
            "mordred:VE1_DzZ",
            "mordred:VE2_DzZ",
            "mordred:VE3_DzZ",
            "mordred:VR1_DzZ",
            "mordred:VR2_DzZ",
            "mordred:VR3_DzZ",
            "mordred:SpAbs_Dzm",
            "mordred:SpMax_Dzm",
            "mordred:SpDiam_Dzm",
            "mordred:SpAD_Dzm",
            "mordred:SpMAD_Dzm",
            "mordred:LogEE_Dzm",
            "mordred:SM1_Dzm",
            "mordred:VE1_Dzm",
            "mordred:VE2_Dzm",
            "mordred:VE3_Dzm",
            "mordred:VR1_Dzm",
            "mordred:VR2_Dzm",
            "mordred:VR3_Dzm",
            "mordred:SpAbs_Dzv",
            "mordred:SpMax_Dzv",
            "mordred:SpDiam_Dzv",
            "mordred:SpAD_Dzv",
            "mordred:SpMAD_Dzv",
            "mordred:LogEE_Dzv",
            "mordred:SM1_Dzv",
            "mordred:VE1_Dzv",
            "mordred:VE2_Dzv",
            "mordred:VE3_Dzv",
            "mordred:VR1_Dzv",
            "mordred:VR2_Dzv",
            "mordred:VR3_Dzv",
            "mordred:SpAbs_Dzse",
            "mordred:SpMax_Dzse",
            "mordred:SpDiam_Dzse",
            "mordred:SpAD_Dzse",
            "mordred:SpMAD_Dzse",
            "mordred:LogEE_Dzse",
            "mordred:SM1_Dzse",
            "mordred:VE1_Dzse",
            "mordred:VE2_Dzse",
            "mordred:VE3_Dzse",
            "mordred:VR1_Dzse",
            "mordred:VR2_Dzse",
            "mordred:VR3_Dzse",
            "mordred:SpAbs_Dzpe",
            "mordred:SpMax_Dzpe",
            "mordred:SpDiam_Dzpe",
            "mordred:SpAD_Dzpe",
            "mordred:SpMAD_Dzpe",
            "mordred:LogEE_Dzpe",
            "mordred:SM1_Dzpe",
            "mordred:VE1_Dzpe",
            "mordred:VE2_Dzpe",
            "mordred:VE3_Dzpe",
            "mordred:VR1_Dzpe",
            "mordred:VR2_Dzpe",
            "mordred:VR3_Dzpe",
            "mordred:SpAbs_Dzare",
            "mordred:SpMax_Dzare",
            "mordred:SpDiam_Dzare",
            "mordred:SpAD_Dzare",
            "mordred:SpMAD_Dzare",
            "mordred:LogEE_Dzare",
            "mordred:SM1_Dzare",
            "mordred:VE1_Dzare",
            "mordred:VE2_Dzare",
            "mordred:VE3_Dzare",
            "mordred:VR1_Dzare",
            "mordred:VR2_Dzare",
            "mordred:VR3_Dzare",
            "mordred:SpAbs_Dzp",
            "mordred:SpMax_Dzp",
            "mordred:SpDiam_Dzp",
            "mordred:SpAD_Dzp",
            "mordred:SpMAD_Dzp",
            "mordred:LogEE_Dzp",
            "mordred:SM1_Dzp",
            "mordred:VE1_Dzp",
            "mordred:VE2_Dzp",
            "mordred:VE3_Dzp",
            "mordred:VR1_Dzp",
            "mordred:VR2_Dzp",
            "mordred:VR3_Dzp",
            "mordred:SpAbs_Dzi",
            "mordred:SpMax_Dzi",
            "mordred:SpDiam_Dzi",
            "mordred:SpAD_Dzi",
            "mordred:SpMAD_Dzi",
            "mordred:LogEE_Dzi",
            "mordred:SM1_Dzi",
            "mordred:VE1_Dzi",
            "mordred:VE2_Dzi",
            "mordred:VE3_Dzi",
            "mordred:VR1_Dzi",
            "mordred:VR2_Dzi",
            "mordred:VR3_Dzi",
            "mordred:BertzCT",
            "mordred:nBonds",
            "mordred:nBondsO",
            "mordred:nBondsS",
            "mordred:nBondsD",
            "mordred:nBondsT",
            "mordred:nBondsA",
            "mordred:nBondsM",
            "mordred:nBondsKS",
            "mordred:nBondsKD",
            "mordred:PNSA1",
            "mordred:PNSA2",
            "mordred:PNSA3",
            "mordred:PNSA4",
            "mordred:PNSA5",
            "mordred:PPSA1",
            "mordred:PPSA2",
            "mordred:PPSA3",
            "mordred:PPSA4",
            "mordred:PPSA5",
            "mordred:DPSA1",
            "mordred:DPSA2",
            "mordred:DPSA3",
            "mordred:DPSA4",
            "mordred:DPSA5",
            "mordred:FNSA1",
            "mordred:FNSA2",
            "mordred:FNSA3",
            "mordred:FNSA4",
            "mordred:FNSA5",
            "mordred:FPSA1",
            "mordred:FPSA2",
            "mordred:FPSA3",
            "mordred:FPSA4",
            "mordred:FPSA5",
            "mordred:WNSA1",
            "mordred:WNSA2",
            "mordred:WNSA3",
            "mordred:WNSA4",
            "mordred:WNSA5",
            "mordred:WPSA1",
            "mordred:WPSA2",
            "mordred:WPSA3",
            "mordred:WPSA4",
            "mordred:WPSA5",
            "mordred:RNCG",
            "mordred:RPCG",
            "mordred:RNCS",
            "mordred:RPCS",
            "mordred:TASA",
            "mordred:TPSA",
            "mordred:RASA",
            "mordred:RPSA",
            "mordred:C1SP1",
            "mordred:C2SP1",
            "mordred:C1SP2",
            "mordred:C2SP2",
            "mordred:C3SP2",
            "mordred:C1SP3",
            "mordred:C2SP3",
            "mordred:C3SP3",
            "mordred:C4SP3",
            "mordred:HybRatio",
            "mordred:FCSP3",
            "mordred:Xch-3d",
            "mordred:Xch-4d",
            "mordred:Xch-5d",
            "mordred:Xch-6d",
            "mordred:Xch-7d",
            "mordred:Xch-3dv",
            "mordred:Xch-4dv",
            "mordred:Xch-5dv",
            "mordred:Xch-6dv",
            "mordred:Xch-7dv",
            "mordred:Xc-3d",
            "mordred:Xc-4d",
            "mordred:Xc-5d",
            "mordred:Xc-6d",
            "mordred:Xc-3dv",
            "mordred:Xc-4dv",
            "mordred:Xc-5dv",
            "mordred:Xc-6dv",
            "mordred:Xpc-4d",
            "mordred:Xpc-5d",
            "mordred:Xpc-6d",
            "mordred:Xpc-4dv",
            "mordred:Xpc-5dv",
            "mordred:Xpc-6dv",
            "mordred:Xp-0d",
            "mordred:Xp-1d",
            "mordred:Xp-2d",
            "mordred:Xp-3d",
            "mordred:Xp-4d",
            "mordred:Xp-5d",
            "mordred:Xp-6d",
            "mordred:Xp-7d",
            "mordred:AXp-0d",
            "mordred:AXp-1d",
            "mordred:AXp-2d",
            "mordred:AXp-3d",
            "mordred:AXp-4d",
            "mordred:AXp-5d",
            "mordred:AXp-6d",
            "mordred:AXp-7d",
            "mordred:Xp-0dv",
            "mordred:Xp-1dv",
            "mordred:Xp-2dv",
            "mordred:Xp-3dv",
            "mordred:Xp-4dv",
            "mordred:Xp-5dv",
            "mordred:Xp-6dv",
            "mordred:Xp-7dv",
            "mordred:AXp-0dv",
            "mordred:AXp-1dv",
            "mordred:AXp-2dv",
            "mordred:AXp-3dv",
            "mordred:AXp-4dv",
            "mordred:AXp-5dv",
            "mordred:AXp-6dv",
            "mordred:AXp-7dv",
            "mordred:SZ",
            "mordred:Sm",
            "mordred:Sv",
            "mordred:Sse",
            "mordred:Spe",
            "mordred:Sare",
            "mordred:Sp",
            "mordred:Si",
            "mordred:MZ",
            "mordred:Mm",
            "mordred:Mv",
            "mordred:Mse",
            "mordred:Mpe",
            "mordred:Mare",
            "mordred:Mp",
            "mordred:Mi",
            "mordred:SpAbs_Dt",
            "mordred:SpMax_Dt",
            "mordred:SpDiam_Dt",
            "mordred:SpAD_Dt",
            "mordred:SpMAD_Dt",
            "mordred:LogEE_Dt",
            "mordred:SM1_Dt",
            "mordred:VE1_Dt",
            "mordred:VE2_Dt",
            "mordred:VE3_Dt",
            "mordred:VR1_Dt",
            "mordred:VR2_Dt",
            "mordred:VR3_Dt",
            "mordred:DetourIndex",
            "mordred:SpAbs_D",
            "mordred:SpMax_D",
            "mordred:SpDiam_D",
            "mordred:SpAD_D",
            "mordred:SpMAD_D",
            "mordred:LogEE_D",
            "mordred:VE1_D",
            "mordred:VE2_D",
            "mordred:VE3_D",
            "mordred:VR1_D",
            "mordred:VR2_D",
            "mordred:VR3_D",
            "mordred:NsLi",
            "mordred:NssBe",
            "mordred:NssssBe",
            "mordred:NssBH",
            "mordred:NsssB",
            "mordred:NssssB",
            "mordred:NsCH3",
            "mordred:NdCH2",
            "mordred:NssCH2",
            "mordred:NtCH",
            "mordred:NdsCH",
            "mordred:NaaCH",
            "mordred:NsssCH",
            "mordred:NddC",
            "mordred:NtsC",
            "mordred:NdssC",
            "mordred:NaasC",
            "mordred:NaaaC",
            "mordred:NssssC",
            "mordred:NsNH3",
            "mordred:NsNH2",
            "mordred:NssNH2",
            "mordred:NdNH",
            "mordred:NssNH",
            "mordred:NaaNH",
            "mordred:NtN",
            "mordred:NsssNH",
            "mordred:NdsN",
            "mordred:NaaN",
            "mordred:NsssN",
            "mordred:NddsN",
            "mordred:NaasN",
            "mordred:NssssN",
            "mordred:NsOH",
            "mordred:NdO",
            "mordred:NssO",
            "mordred:NaaO",
            "mordred:NsF",
            "mordred:NsSiH3",
            "mordred:NssSiH2",
            "mordred:NsssSiH",
            "mordred:NssssSi",
            "mordred:NsPH2",
            "mordred:NssPH",
            "mordred:NsssP",
            "mordred:NdsssP",
            "mordred:NsssssP",
            "mordred:NsSH",
            "mordred:NdS",
            "mordred:NssS",
            "mordred:NaaS",
            "mordred:NdssS",
            "mordred:NddssS",
            "mordred:NsCl",
            "mordred:NsGeH3",
            "mordred:NssGeH2",
            "mordred:NsssGeH",
            "mordred:NssssGe",
            "mordred:NsAsH2",
            "mordred:NssAsH",
            "mordred:NsssAs",
            "mordred:NsssdAs",
            "mordred:NsssssAs",
            "mordred:NsSeH",
            "mordred:NdSe",
            "mordred:NssSe",
            "mordred:NaaSe",
            "mordred:NdssSe",
            "mordred:NddssSe",
            "mordred:NsBr",
            "mordred:NsSnH3",
            "mordred:NssSnH2",
            "mordred:NsssSnH",
            "mordred:NssssSn",
            "mordred:NsI",
            "mordred:NsPbH3",
            "mordred:NssPbH2",
            "mordred:NsssPbH",
            "mordred:NssssPb",
            "mordred:SsLi",
            "mordred:SssBe",
            "mordred:SssssBe",
            "mordred:SssBH",
            "mordred:SsssB",
            "mordred:SssssB",
            "mordred:SsCH3",
            "mordred:SdCH2",
            "mordred:SssCH2",
            "mordred:StCH",
            "mordred:SdsCH",
            "mordred:SaaCH",
            "mordred:SsssCH",
            "mordred:SddC",
            "mordred:StsC",
            "mordred:SdssC",
            "mordred:SaasC",
            "mordred:SaaaC",
            "mordred:SssssC",
            "mordred:SsNH3",
            "mordred:SsNH2",
            "mordred:SssNH2",
            "mordred:SdNH",
            "mordred:SssNH",
            "mordred:SaaNH",
            "mordred:StN",
            "mordred:SsssNH",
            "mordred:SdsN",
            "mordred:SaaN",
            "mordred:SsssN",
            "mordred:SddsN",
            "mordred:SaasN",
            "mordred:SssssN",
            "mordred:SsOH",
            "mordred:SdO",
            "mordred:SssO",
            "mordred:SaaO",
            "mordred:SsF",
            "mordred:SsSiH3",
            "mordred:SssSiH2",
            "mordred:SsssSiH",
            "mordred:SssssSi",
            "mordred:SsPH2",
            "mordred:SssPH",
            "mordred:SsssP",
            "mordred:SdsssP",
            "mordred:SsssssP",
            "mordred:SsSH",
            "mordred:SdS",
            "mordred:SssS",
            "mordred:SaaS",
            "mordred:SdssS",
            "mordred:SddssS",
            "mordred:SsCl",
            "mordred:SsGeH3",
            "mordred:SssGeH2",
            "mordred:SsssGeH",
            "mordred:SssssGe",
            "mordred:SsAsH2",
            "mordred:SssAsH",
            "mordred:SsssAs",
            "mordred:SsssdAs",
            "mordred:SsssssAs",
            "mordred:SsSeH",
            "mordred:SdSe",
            "mordred:SssSe",
            "mordred:SaaSe",
            "mordred:SdssSe",
            "mordred:SddssSe",
            "mordred:SsBr",
            "mordred:SsSnH3",
            "mordred:SssSnH2",
            "mordred:SsssSnH",
            "mordred:SssssSn",
            "mordred:SsI",
            "mordred:SsPbH3",
            "mordred:SssPbH2",
            "mordred:SsssPbH",
            "mordred:SssssPb",
            "mordred:MAXsLi",
            "mordred:MAXssBe",
            "mordred:MAXssssBe",
            "mordred:MAXssBH",
            "mordred:MAXsssB",
            "mordred:MAXssssB",
            "mordred:MAXsCH3",
            "mordred:MAXdCH2",
            "mordred:MAXssCH2",
            "mordred:MAXtCH",
            "mordred:MAXdsCH",
            "mordred:MAXaaCH",
            "mordred:MAXsssCH",
            "mordred:MAXddC",
            "mordred:MAXtsC",
            "mordred:MAXdssC",
            "mordred:MAXaasC",
            "mordred:MAXaaaC",
            "mordred:MAXssssC",
            "mordred:MAXsNH3",
            "mordred:MAXsNH2",
            "mordred:MAXssNH2",
            "mordred:MAXdNH",
            "mordred:MAXssNH",
            "mordred:MAXaaNH",
            "mordred:MAXtN",
            "mordred:MAXsssNH",
            "mordred:MAXdsN",
            "mordred:MAXaaN",
            "mordred:MAXsssN",
            "mordred:MAXddsN",
            "mordred:MAXaasN",
            "mordred:MAXssssN",
            "mordred:MAXsOH",
            "mordred:MAXdO",
            "mordred:MAXssO",
            "mordred:MAXaaO",
            "mordred:MAXsF",
            "mordred:MAXsSiH3",
            "mordred:MAXssSiH2",
            "mordred:MAXsssSiH",
            "mordred:MAXssssSi",
            "mordred:MAXsPH2",
            "mordred:MAXssPH",
            "mordred:MAXsssP",
            "mordred:MAXdsssP",
            "mordred:MAXsssssP",
            "mordred:MAXsSH",
            "mordred:MAXdS",
            "mordred:MAXssS",
            "mordred:MAXaaS",
            "mordred:MAXdssS",
            "mordred:MAXddssS",
            "mordred:MAXsCl",
            "mordred:MAXsGeH3",
            "mordred:MAXssGeH2",
            "mordred:MAXsssGeH",
            "mordred:MAXssssGe",
            "mordred:MAXsAsH2",
            "mordred:MAXssAsH",
            "mordred:MAXsssAs",
            "mordred:MAXsssdAs",
            "mordred:MAXsssssAs",
            "mordred:MAXsSeH",
            "mordred:MAXdSe",
            "mordred:MAXssSe",
            "mordred:MAXaaSe",
            "mordred:MAXdssSe",
            "mordred:MAXddssSe",
            "mordred:MAXsBr",
            "mordred:MAXsSnH3",
            "mordred:MAXssSnH2",
            "mordred:MAXsssSnH",
            "mordred:MAXssssSn",
            "mordred:MAXsI",
            "mordred:MAXsPbH3",
            "mordred:MAXssPbH2",
            "mordred:MAXsssPbH",
            "mordred:MAXssssPb",
            "mordred:MINsLi",
            "mordred:MINssBe",
            "mordred:MINssssBe",
            "mordred:MINssBH",
            "mordred:MINsssB",
            "mordred:MINssssB",
            "mordred:MINsCH3",
            "mordred:MINdCH2",
            "mordred:MINssCH2",
            "mordred:MINtCH",
            "mordred:MINdsCH",
            "mordred:MINaaCH",
            "mordred:MINsssCH",
            "mordred:MINddC",
            "mordred:MINtsC",
            "mordred:MINdssC",
            "mordred:MINaasC",
            "mordred:MINaaaC",
            "mordred:MINssssC",
            "mordred:MINsNH3",
            "mordred:MINsNH2",
            "mordred:MINssNH2",
            "mordred:MINdNH",
            "mordred:MINssNH",
            "mordred:MINaaNH",
            "mordred:MINtN",
            "mordred:MINsssNH",
            "mordred:MINdsN",
            "mordred:MINaaN",
            "mordred:MINsssN",
            "mordred:MINddsN",
            "mordred:MINaasN",
            "mordred:MINssssN",
            "mordred:MINsOH",
            "mordred:MINdO",
            "mordred:MINssO",
            "mordred:MINaaO",
            "mordred:MINsF",
            "mordred:MINsSiH3",
            "mordred:MINssSiH2",
            "mordred:MINsssSiH",
            "mordred:MINssssSi",
            "mordred:MINsPH2",
            "mordred:MINssPH",
            "mordred:MINsssP",
            "mordred:MINdsssP",
            "mordred:MINsssssP",
            "mordred:MINsSH",
            "mordred:MINdS",
            "mordred:MINssS",
            "mordred:MINaaS",
            "mordred:MINdssS",
            "mordred:MINddssS",
            "mordred:MINsCl",
            "mordred:MINsGeH3",
            "mordred:MINssGeH2",
            "mordred:MINsssGeH",
            "mordred:MINssssGe",
            "mordred:MINsAsH2",
            "mordred:MINssAsH",
            "mordred:MINsssAs",
            "mordred:MINsssdAs",
            "mordred:MINsssssAs",
            "mordred:MINsSeH",
            "mordred:MINdSe",
            "mordred:MINssSe",
            "mordred:MINaaSe",
            "mordred:MINdssSe",
            "mordred:MINddssSe",
            "mordred:MINsBr",
            "mordred:MINsSnH3",
            "mordred:MINssSnH2",
            "mordred:MINsssSnH",
            "mordred:MINssssSn",
            "mordred:MINsI",
            "mordred:MINsPbH3",
            "mordred:MINssPbH2",
            "mordred:MINsssPbH",
            "mordred:MINssssPb",
            "mordred:ECIndex",
            "mordred:ETA_alpha",
            "mordred:AETA_alpha",
            "mordred:ETA_shape_p",
            "mordred:ETA_shape_y",
            "mordred:ETA_shape_x",
            "mordred:ETA_beta",
            "mordred:AETA_beta",
            "mordred:ETA_beta_s",
            "mordred:AETA_beta_s",
            "mordred:ETA_beta_ns",
            "mordred:AETA_beta_ns",
            "mordred:ETA_beta_ns_d",
            "mordred:AETA_beta_ns_d",
            "mordred:ETA_eta",
            "mordred:AETA_eta",
            "mordred:ETA_eta_L",
            "mordred:AETA_eta_L",
            "mordred:ETA_eta_R",
            "mordred:AETA_eta_R",
            "mordred:ETA_eta_RL",
            "mordred:AETA_eta_RL",
            "mordred:ETA_eta_F",
            "mordred:AETA_eta_F",
            "mordred:ETA_eta_FL",
            "mordred:AETA_eta_FL",
            "mordred:ETA_eta_B",
            "mordred:AETA_eta_B",
            "mordred:ETA_eta_BR",
            "mordred:AETA_eta_BR",
            "mordred:ETA_dAlpha_A",
            "mordred:ETA_dAlpha_B",
            "mordred:ETA_epsilon_1",
            "mordred:ETA_epsilon_2",
            "mordred:ETA_epsilon_3",
            "mordred:ETA_epsilon_4",
            "mordred:ETA_epsilon_5",
            "mordred:ETA_dEpsilon_A",
            "mordred:ETA_dEpsilon_B",
            "mordred:ETA_dEpsilon_C",
            "mordred:ETA_dEpsilon_D",
            "mordred:ETA_dBeta",
            "mordred:AETA_dBeta",
            "mordred:ETA_psi_1",
            "mordred:ETA_dPsi_A",
            "mordred:ETA_dPsi_B",
            "mordred:fragCpx",
            "mordred:fMF",
            "mordred:GeomDiameter",
            "mordred:GeomRadius",
            "mordred:GeomShapeIndex",
            "mordred:GeomPetitjeanIndex",
            "mordred:GRAV",
            "mordred:GRAVH",
            "mordred:GRAVp",
            "mordred:GRAVHp",
            "mordred:nHBAcc",
            "mordred:nHBDon",
            "mordred:IC0",
            "mordred:IC1",
            "mordred:IC2",
            "mordred:IC3",
            "mordred:IC4",
            "mordred:IC5",
            "mordred:TIC0",
            "mordred:TIC1",
            "mordred:TIC2",
            "mordred:TIC3",
            "mordred:TIC4",
            "mordred:TIC5",
            "mordred:SIC0",
            "mordred:SIC1",
            "mordred:SIC2",
            "mordred:SIC3",
            "mordred:SIC4",
            "mordred:SIC5",
            "mordred:BIC0",
            "mordred:BIC1",
            "mordred:BIC2",
            "mordred:BIC3",
            "mordred:BIC4",
            "mordred:BIC5",
            "mordred:CIC0",
            "mordred:CIC1",
            "mordred:CIC2",
            "mordred:CIC3",
            "mordred:CIC4",
            "mordred:CIC5",
            "mordred:MIC0",
            "mordred:MIC1",
            "mordred:MIC2",
            "mordred:MIC3",
            "mordred:MIC4",
            "mordred:MIC5",
            "mordred:ZMIC0",
            "mordred:ZMIC1",
            "mordred:ZMIC2",
            "mordred:ZMIC3",
            "mordred:ZMIC4",
            "mordred:ZMIC5",
            "mordred:Kier1",
            "mordred:Kier2",
            "mordred:Kier3",
            "mordred:Lipinski",
            "mordred:GhoseFilter",
            "mordred:FilterItLogS",
            "mordred:VMcGowan",
            "mordred:Mor01",
            "mordred:Mor02",
            "mordred:Mor03",
            "mordred:Mor04",
            "mordred:Mor05",
            "mordred:Mor06",
            "mordred:Mor07",
            "mordred:Mor08",
            "mordred:Mor09",
            "mordred:Mor10",
            "mordred:Mor11",
            "mordred:Mor12",
            "mordred:Mor13",
            "mordred:Mor14",
            "mordred:Mor15",
            "mordred:Mor16",
            "mordred:Mor17",
            "mordred:Mor18",
            "mordred:Mor19",
            "mordred:Mor20",
            "mordred:Mor21",
            "mordred:Mor22",
            "mordred:Mor23",
            "mordred:Mor24",
            "mordred:Mor25",
            "mordred:Mor26",
            "mordred:Mor27",
            "mordred:Mor28",
            "mordred:Mor29",
            "mordred:Mor30",
            "mordred:Mor31",
            "mordred:Mor32",
            "mordred:Mor01m",
            "mordred:Mor02m",
            "mordred:Mor03m",
            "mordred:Mor04m",
            "mordred:Mor05m",
            "mordred:Mor06m",
            "mordred:Mor07m",
            "mordred:Mor08m",
            "mordred:Mor09m",
            "mordred:Mor10m",
            "mordred:Mor11m",
            "mordred:Mor12m",
            "mordred:Mor13m",
            "mordred:Mor14m",
            "mordred:Mor15m",
            "mordred:Mor16m",
            "mordred:Mor17m",
            "mordred:Mor18m",
            "mordred:Mor19m",
            "mordred:Mor20m",
            "mordred:Mor21m",
            "mordred:Mor22m",
            "mordred:Mor23m",
            "mordred:Mor24m",
            "mordred:Mor25m",
            "mordred:Mor26m",
            "mordred:Mor27m",
            "mordred:Mor28m",
            "mordred:Mor29m",
            "mordred:Mor30m",
            "mordred:Mor31m",
            "mordred:Mor32m",
            "mordred:Mor01v",
            "mordred:Mor02v",
            "mordred:Mor03v",
            "mordred:Mor04v",
            "mordred:Mor05v",
            "mordred:Mor06v",
            "mordred:Mor07v",
            "mordred:Mor08v",
            "mordred:Mor09v",
            "mordred:Mor10v",
            "mordred:Mor11v",
            "mordred:Mor12v",
            "mordred:Mor13v",
            "mordred:Mor14v",
            "mordred:Mor15v",
            "mordred:Mor16v",
            "mordred:Mor17v",
            "mordred:Mor18v",
            "mordred:Mor19v",
            "mordred:Mor20v",
            "mordred:Mor21v",
            "mordred:Mor22v",
            "mordred:Mor23v",
            "mordred:Mor24v",
            "mordred:Mor25v",
            "mordred:Mor26v",
            "mordred:Mor27v",
            "mordred:Mor28v",
            "mordred:Mor29v",
            "mordred:Mor30v",
            "mordred:Mor31v",
            "mordred:Mor32v",
            "mordred:Mor01se",
            "mordred:Mor02se",
            "mordred:Mor03se",
            "mordred:Mor04se",
            "mordred:Mor05se",
            "mordred:Mor06se",
            "mordred:Mor07se",
            "mordred:Mor08se",
            "mordred:Mor09se",
            "mordred:Mor10se",
            "mordred:Mor11se",
            "mordred:Mor12se",
            "mordred:Mor13se",
            "mordred:Mor14se",
            "mordred:Mor15se",
            "mordred:Mor16se",
            "mordred:Mor17se",
            "mordred:Mor18se",
            "mordred:Mor19se",
            "mordred:Mor20se",
            "mordred:Mor21se",
            "mordred:Mor22se",
            "mordred:Mor23se",
            "mordred:Mor24se",
            "mordred:Mor25se",
            "mordred:Mor26se",
            "mordred:Mor27se",
            "mordred:Mor28se",
            "mordred:Mor29se",
            "mordred:Mor30se",
            "mordred:Mor31se",
            "mordred:Mor32se",
            "mordred:Mor01p",
            "mordred:Mor02p",
            "mordred:Mor03p",
            "mordred:Mor04p",
            "mordred:Mor05p",
            "mordred:Mor06p",
            "mordred:Mor07p",
            "mordred:Mor08p",
            "mordred:Mor09p",
            "mordred:Mor10p",
            "mordred:Mor11p",
            "mordred:Mor12p",
            "mordred:Mor13p",
            "mordred:Mor14p",
            "mordred:Mor15p",
            "mordred:Mor16p",
            "mordred:Mor17p",
            "mordred:Mor18p",
            "mordred:Mor19p",
            "mordred:Mor20p",
            "mordred:Mor21p",
            "mordred:Mor22p",
            "mordred:Mor23p",
            "mordred:Mor24p",
            "mordred:Mor25p",
            "mordred:Mor26p",
            "mordred:Mor27p",
            "mordred:Mor28p",
            "mordred:Mor29p",
            "mordred:Mor30p",
            "mordred:Mor31p",
            "mordred:Mor32p",
            "mordred:LabuteASA",
            "mordred:PEOE_VSA1",
            "mordred:PEOE_VSA2",
            "mordred:PEOE_VSA3",
            "mordred:PEOE_VSA4",
            "mordred:PEOE_VSA5",
            "mordred:PEOE_VSA6",
            "mordred:PEOE_VSA7",
            "mordred:PEOE_VSA8",
            "mordred:PEOE_VSA9",
            "mordred:PEOE_VSA10",
            "mordred:PEOE_VSA11",
            "mordred:PEOE_VSA12",
            "mordred:PEOE_VSA13",
            "mordred:SMR_VSA1",
            "mordred:SMR_VSA2",
            "mordred:SMR_VSA3",
            "mordred:SMR_VSA4",
            "mordred:SMR_VSA5",
            "mordred:SMR_VSA6",
            "mordred:SMR_VSA7",
            "mordred:SMR_VSA8",
            "mordred:SMR_VSA9",
            "mordred:SlogP_VSA1",
            "mordred:SlogP_VSA2",
            "mordred:SlogP_VSA3",
            "mordred:SlogP_VSA4",
            "mordred:SlogP_VSA5",
            "mordred:SlogP_VSA6",
            "mordred:SlogP_VSA7",
            "mordred:SlogP_VSA8",
            "mordred:SlogP_VSA9",
            "mordred:SlogP_VSA10",
            "mordred:SlogP_VSA11",
            "mordred:EState_VSA1",
            "mordred:EState_VSA2",
            "mordred:EState_VSA3",
            "mordred:EState_VSA4",
            "mordred:EState_VSA5",
            "mordred:EState_VSA6",
            "mordred:EState_VSA7",
            "mordred:EState_VSA8",
            "mordred:EState_VSA9",
            "mordred:EState_VSA10",
            "mordred:VSA_EState1",
            "mordred:VSA_EState2",
            "mordred:VSA_EState3",
            "mordred:VSA_EState4",
            "mordred:VSA_EState5",
            "mordred:VSA_EState6",
            "mordred:VSA_EState7",
            "mordred:VSA_EState8",
            "mordred:VSA_EState9",
            "mordred:MDEC-11",
            "mordred:MDEC-12",
            "mordred:MDEC-13",
            "mordred:MDEC-14",
            "mordred:MDEC-22",
            "mordred:MDEC-23",
            "mordred:MDEC-24",
            "mordred:MDEC-33",
            "mordred:MDEC-34",
            "mordred:MDEC-44",
            "mordred:MDEO-11",
            "mordred:MDEO-12",
            "mordred:MDEO-22",
            "mordred:MDEN-11",
            "mordred:MDEN-12",
            "mordred:MDEN-13",
            "mordred:MDEN-22",
            "mordred:MDEN-23",
            "mordred:MDEN-33",
            "mordred:MID",
            "mordred:AMID",
            "mordred:MID_h",
            "mordred:AMID_h",
            "mordred:MID_C",
            "mordred:AMID_C",
            "mordred:MID_N",
            "mordred:AMID_N",
            "mordred:MID_O",
            "mordred:AMID_O",
            "mordred:MID_X",
            "mordred:AMID_X",
            "mordred:MOMI-X",
            "mordred:MOMI-Y",
            "mordred:MOMI-Z",
            "mordred:PBF",
            "mordred:MPC2",
            "mordred:MPC3",
            "mordred:MPC4",
            "mordred:MPC5",
            "mordred:MPC6",
            "mordred:MPC7",
            "mordred:MPC8",
            "mordred:MPC9",
            "mordred:MPC10",
            "mordred:TMPC10",
            "mordred:piPC1",
            "mordred:piPC2",
            "mordred:piPC3",
            "mordred:piPC4",
            "mordred:piPC5",
            "mordred:piPC6",
            "mordred:piPC7",
            "mordred:piPC8",
            "mordred:piPC9",
            "mordred:piPC10",
            "mordred:TpiPC10",
            "mordred:apol",
            "mordred:bpol",
            "mordred:nRing",
            "mordred:n3Ring",
            "mordred:n4Ring",
            "mordred:n5Ring",
            "mordred:n6Ring",
            "mordred:n7Ring",
            "mordred:n8Ring",
            "mordred:n9Ring",
            "mordred:n10Ring",
            "mordred:n11Ring",
            "mordred:n12Ring",
            "mordred:nG12Ring",
            "mordred:nHRing",
            "mordred:n3HRing",
            "mordred:n4HRing",
            "mordred:n5HRing",
            "mordred:n6HRing",
            "mordred:n7HRing",
            "mordred:n8HRing",
            "mordred:n9HRing",
            "mordred:n10HRing",
            "mordred:n11HRing",
            "mordred:n12HRing",
            "mordred:nG12HRing",
            "mordred:naRing",
            "mordred:n3aRing",
            "mordred:n4aRing",
            "mordred:n5aRing",
            "mordred:n6aRing",
            "mordred:n7aRing",
            "mordred:n8aRing",
            "mordred:n9aRing",
            "mordred:n10aRing",
            "mordred:n11aRing",
            "mordred:n12aRing",
            "mordred:nG12aRing",
            "mordred:naHRing",
            "mordred:n3aHRing",
            "mordred:n4aHRing",
            "mordred:n5aHRing",
            "mordred:n6aHRing",
            "mordred:n7aHRing",
            "mordred:n8aHRing",
            "mordred:n9aHRing",
            "mordred:n10aHRing",
            "mordred:n11aHRing",
            "mordred:n12aHRing",
            "mordred:nG12aHRing",
            "mordred:nARing",
            "mordred:n3ARing",
            "mordred:n4ARing",
            "mordred:n5ARing",
            "mordred:n6ARing",
            "mordred:n7ARing",
            "mordred:n8ARing",
            "mordred:n9ARing",
            "mordred:n10ARing",
            "mordred:n11ARing",
            "mordred:n12ARing",
            "mordred:nG12ARing",
            "mordred:nAHRing",
            "mordred:n3AHRing",
            "mordred:n4AHRing",
            "mordred:n5AHRing",
            "mordred:n6AHRing",
            "mordred:n7AHRing",
            "mordred:n8AHRing",
            "mordred:n9AHRing",
            "mordred:n10AHRing",
            "mordred:n11AHRing",
            "mordred:n12AHRing",
            "mordred:nG12AHRing",
            "mordred:nFRing",
            "mordred:n4FRing",
            "mordred:n5FRing",
            "mordred:n6FRing",
            "mordred:n7FRing",
            "mordred:n8FRing",
            "mordred:n9FRing",
            "mordred:n10FRing",
            "mordred:n11FRing",
            "mordred:n12FRing",
            "mordred:nG12FRing",
            "mordred:nFHRing",
            "mordred:n4FHRing",
            "mordred:n5FHRing",
            "mordred:n6FHRing",
            "mordred:n7FHRing",
            "mordred:n8FHRing",
            "mordred:n9FHRing",
            "mordred:n10FHRing",
            "mordred:n11FHRing",
            "mordred:n12FHRing",
            "mordred:nG12FHRing",
            "mordred:nFaRing",
            "mordred:n4FaRing",
            "mordred:n5FaRing",
            "mordred:n6FaRing",
            "mordred:n7FaRing",
            "mordred:n8FaRing",
            "mordred:n9FaRing",
            "mordred:n10FaRing",
            "mordred:n11FaRing",
            "mordred:n12FaRing",
            "mordred:nG12FaRing",
            "mordred:nFaHRing",
            "mordred:n4FaHRing",
            "mordred:n5FaHRing",
            "mordred:n6FaHRing",
            "mordred:n7FaHRing",
            "mordred:n8FaHRing",
            "mordred:n9FaHRing",
            "mordred:n10FaHRing",
            "mordred:n11FaHRing",
            "mordred:n12FaHRing",
            "mordred:nG12FaHRing",
            "mordred:nFARing",
            "mordred:n4FARing",
            "mordred:n5FARing",
            "mordred:n6FARing",
            "mordred:n7FARing",
            "mordred:n8FARing",
            "mordred:n9FARing",
            "mordred:n10FARing",
            "mordred:n11FARing",
            "mordred:n12FARing",
            "mordred:nG12FARing",
            "mordred:nFAHRing",
            "mordred:n4FAHRing",
            "mordred:n5FAHRing",
            "mordred:n6FAHRing",
            "mordred:n7FAHRing",
            "mordred:n8FAHRing",
            "mordred:n9FAHRing",
            "mordred:n10FAHRing",
            "mordred:n11FAHRing",
            "mordred:n12FAHRing",
            "mordred:nG12FAHRing",
            "mordred:nRot",
            "mordred:RotRatio",
            "mordred:SLogP",
            "mordred:SMR",
            "mordred:TopoPSA(NO)",
            "mordred:TopoPSA",
            "mordred:GGI1",
            "mordred:GGI2",
            "mordred:GGI3",
            "mordred:GGI4",
            "mordred:GGI5",
            "mordred:GGI6",
            "mordred:GGI7",
            "mordred:GGI8",
            "mordred:GGI9",
            "mordred:GGI10",
            "mordred:JGI1",
            "mordred:JGI2",
            "mordred:JGI3",
            "mordred:JGI4",
            "mordred:JGI5",
            "mordred:JGI6",
            "mordred:JGI7",
            "mordred:JGI8",
            "mordred:JGI9",
            "mordred:JGI10",
            "mordred:JGT10",
            "mordred:Diameter",
            "mordred:Radius",
            "mordred:TopoShapeIndex",
            "mordred:PetitjeanIndex",
            "mordred:Vabc",
            "mordred:VAdjMat",
            "mordred:MWC01",
            "mordred:MWC02",
            "mordred:MWC03",
            "mordred:MWC04",
            "mordred:MWC05",
            "mordred:MWC06",
            "mordred:MWC07",
            "mordred:MWC08",
            "mordred:MWC09",
            "mordred:MWC10",
            "mordred:TMWC10",
            "mordred:SRW02",
            "mordred:SRW03",
            "mordred:SRW04",
            "mordred:SRW05",
            "mordred:SRW06",
            "mordred:SRW07",
            "mordred:SRW08",
            "mordred:SRW09",
            "mordred:SRW10",
            "mordred:TSRW10",
            "mordred:MW",
            "mordred:AMW",
            "mordred:WPath",
            "mordred:WPol",
            "mordred:Zagreb1",
            "mordred:Zagreb2",
            "mordred:mZagreb1",
            "mordred:mZagreb2",
        ]
